/*
 *  $Id: pvrusb2-base.h 496 2005-07-16 19:47:53Z isely $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __PVRUSB2_BASE_H
#define __PVRUSB2_BASE_H

#include <asm/semaphore.h>
#include <linux/usb.h>
#include <linux/workqueue.h>
#include "pvrusb2-hdw.h"

struct pvr2_hdw;     /* hardware interface - defined elsewhere */
struct pvr2_stream;  /* stream interface - defined elsewhere */

struct pvr2_main;        /* All central hardware state */
struct pvr2_channel;     /* One I/O pathway to a user */
struct pvr2_main_stream; /* Wrapper for a stream */
struct pvr2_crit_reg;    /* Critical region pointer */

struct pvr2_main_stream {
    struct pvr2_channel *user;
    struct pvr2_stream *stream;
};

struct pvr2_main {
    struct pvr2_channel *mc_first;
    struct pvr2_channel *mc_last;
    struct pvr2_hdw *hdw;
    struct pvr2_main_stream video_stream;
    struct semaphore sem;
    int disconnect_flag;
    int kthread_ref_count;

    /* Called after pvr2_main initialization is complete */
    void (*setup_func)(struct pvr2_main *);

    /* Work queue overhead for out-of-line processing */
    struct workqueue_struct *workqueue;
    struct work_struct workinit;

};

struct pvr2_channel {
    struct pvr2_main *mc_head;
    struct pvr2_channel *mc_next;
    struct pvr2_channel *mc_prev;
    struct pvr2_main_stream *stream;
    struct pvr2_hdw *hdw;
    void (*check_func)(struct pvr2_channel *);
};

void pvr2_main_enter(struct pvr2_main *);
void pvr2_main_exit(struct pvr2_main *);

struct pvr2_main *pvr2_main_create(struct usb_interface *intf,
                                   void (*setup_func)(struct pvr2_main *));
void pvr2_main_disconnect(struct pvr2_main *);

void pvr2_channel_init(struct pvr2_channel *,struct pvr2_main *);
void pvr2_channel_done(struct pvr2_channel *);
int pvr2_channel_claim_stream(struct pvr2_channel *,
			      struct pvr2_main_stream *);


#endif /* __PVRUSB2_BASE_H */
/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
