/*
 *
 *  $Id: pvrusb2-hdw-low.h 496 2005-07-16 19:47:53Z isely $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __PVRUSB2_HDW_LOW_H
#define __PVRUSB2_HDW_LOW_H

#include <linux/types.h>
#include "pvrusb2-hdw.h"

#define PVR2_DECOMPOSE_LE(t,i,d) \
    do {    \
        (t)[i] = (d) & 0xff;\
        (t)[i+1] = ((d) >> 8) & 0xff;\
        (t)[i+2] = ((d) >> 16) & 0xff;\
        (t)[i+3] = ((d) >> 24) & 0xff;\
    } while(0)

#define PVR2_DECOMPOSE_BE(t,i,d) \
    do {    \
        (t)[i+3] = (d) & 0xff;\
        (t)[i+2] = ((d) >> 8) & 0xff;\
        (t)[i+1] = ((d) >> 16) & 0xff;\
        (t)[i] = ((d) >> 24) & 0xff;\
    } while(0)

#define PVR2_COMPOSE_LE(t,i) \
    ((((u32)((t)[i+3])) << 24) | \
     (((u32)((t)[i+2])) << 16) | \
     (((u32)((t)[i+1])) << 8) | \
     ((u32)((t)[i])))

#define PVR2_COMPOSE_BE(t,i) \
    ((((u32)((t)[i])) << 24) | \
     (((u32)((t)[i+1])) << 16) | \
     (((u32)((t)[i+2])) << 8) | \
     ((u32)((t)[i+3])))

#define PVR2_SIZEOF_CMD(c)   (2*sizeof(short)+sizeof(char)*(c)->length)
#define PVR2_CMD(r,c)   { .length = sizeof(c)-1, .ret = r , .cmd = c }
#define PVR2_DECLARE_CMD(name,r,c) struct pvr2_cmd name = PVR2_CMD(r,c)

struct pvr2_cmd {
    u8 length;
    u8 ret;
    u8 *cmd;
};

void pvr2_reset_ctl_endpoints(struct pvr2_hdw *hdw);

int pvr2_send_sequence(struct pvr2_hdw *, struct pvr2_cmd *);
int pvr2_send_request(struct pvr2_hdw *, void *, unsigned int,
		      void *, unsigned int);
int pvr2_do_command(struct pvr2_hdw *, struct pvr2_cmd *);

int pvr2_write_register(struct pvr2_hdw *, u16, u32);
int pvr2_read_register(struct pvr2_hdw *, u16, u32 *);
int pvr2_write_u16(struct pvr2_hdw *, u16, int);
int pvr2_write_u8(struct pvr2_hdw *, u8, int);

/* Call if for any reason we can't talk to the hardware anymore */
void pvr2_hdw_render_useless(struct pvr2_hdw *);
void pvr2_hdw_render_useless_unlocked(struct pvr2_hdw *);

/* Set / clear 8051's reset bit */
void pvr2_hdw_cpureset_assert(struct pvr2_hdw *,int);

/* Execute a USB-commanded device reset */
void pvr2_hdw_device_reset(struct pvr2_hdw *);

/* Execute hard reset command (after this point it's likely that the
   encoder will have to be reconfigured).  This also clears the "useless"
   state. */
int pvr2_hdw_cmd_deep_reset(struct pvr2_hdw *);

/* Execute simple reset command */
int pvr2_hdw_cmd_soft_reset(struct pvr2_hdw *);

/* Stop / start video stream transport */
int pvr2_hdw_cmd_usbstream(struct pvr2_hdw *hdw,int runFl);

/* Direct manipulation of GPIO bits */
int pvr2_hdw_gpio_get_dir(struct pvr2_hdw *hdw,u32 *);
int pvr2_hdw_gpio_get_out(struct pvr2_hdw *hdw,u32 *);
int pvr2_hdw_gpio_get_in(struct pvr2_hdw *hdw,u32 *);
int pvr2_hdw_gpio_chg_dir(struct pvr2_hdw *hdw,u32 msk,u32 val);
int pvr2_hdw_gpio_chg_out(struct pvr2_hdw *hdw,u32 msk,u32 val);

struct pvr2_hdw_debug_info {
    int big_lock_held;
    int ctl_lock_held;
    int flag_ok;
    int flag_disconnected;
    int flag_init_ok;
    int flag_streaming_enabled;
    unsigned long subsys_flags;
    int cmd_debug_state;
    int cmd_debug_write_len;
    int cmd_debug_read_len;
    int cmd_debug_write_pend;
    int cmd_debug_read_pend;
    int cmd_debug_timeout;
    int cmd_debug_rstatus;
    int cmd_debug_wstatus;
    unsigned char cmd_code;
};

/* Non-intrusively retrieve internal state info - this is useful for
   diagnosing lockups.  Note that this operation is completed without any
   kind of locking and so it is not atomic and may yield inconsistent
   results.  This is *purely* a debugging aid. */
void pvr2_hdw_get_debug_info(const struct pvr2_hdw *hdw,
                             struct pvr2_hdw_debug_info *);

#endif /* __PVRUSB2_HDW_LOW_H */
/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
